from __future__ import annotations

import ast
import os
import typing as t

import tomli
import typer

from pyaphid.analyzer import ExpandedCallCollector, Transformer, Visitor
from pyaphid.helpers import echo_with_line_ref


def pyproject():
    with open("pyproject.toml", "rb") as f:
        return tomli.load(f)


def get_pyaphid_toml_section() -> dict[str, t.Any] | None:
    pyproject_toml = pyproject()
    if "tool" in pyproject_toml:
        if "pyaphid" in pyproject_toml["tool"]:
            return pyproject_toml["tool"]["pyaphid"]
    return None


def get_forbidden_calls() -> list[str]:
    """Ger forbidden calls from pyproject.toml"""
    aphid_section = get_pyaphid_toml_section()
    if aphid_section and "forbidden" in aphid_section:
        return aphid_section["forbidden"]

    return []


def collect_python_files(files_and_dirs: list[str]):
    files: list[str] = []
    for file_or_dir in files_and_dirs:
        path = os.path.abspath(file_or_dir)
        if os.path.isfile(path) and path.endswith(".py"):
            files.append(file_or_dir)
        elif os.path.isdir(path):
            for dirpath, _, filenames in os.walk(file_or_dir, followlinks=False):
                for filename in filenames:
                    if filename.endswith(".py"):
                        files.append(f"{dirpath}/{filename}")

    return files


def main(
    files_and_dirs: list[str] = typer.Argument(
        None, show_default=False, help="Multiple files and/or directories to work with"
    ),
    # omit: bool = typer.Option(
    #    False,
    #   "--omit",
    #   "-o",
    #  rich_help_panel="Pyaphid options",
    # show_default=False,
    # help="Remove calls from falls (might destroy formatting)",
    # ),
    print_names_only: bool = typer.Option(
        False,
        "--names",
        "-n",
        rich_help_panel="Pyaphid options",
        show_default=False,
        help="Print only the expandend names of calls in files",
    ),
):
    files = collect_python_files(files_and_dirs)
    forbidden = get_forbidden_calls()
    omit = False
    exit_code = 0
    if print_names_only:
        cls = ExpandedCallCollector
    elif omit:
        cls = Transformer
    else:
        cls = Visitor

    for filepath in files:
        with open(filepath, "rb") as f:
            tree = ast.parse(f.read())
        visitor = cls(filepath=filepath, forbidden=forbidden)
        visitor.visit(tree)
        if isinstance(visitor, ExpandedCallCollector):
            names = visitor.calls
        else:
            names = visitor.matches
        exit_code += len(names)
        for match, expanded_call_name in names:
            echo_with_line_ref(filepath, match, expanded_call_name)

        # if omit and names:
        #    with open(filepath, "w") as f:
        #        f.write(ast.unparse(tree))

    raise typer.Exit(exit_code if not print_names_only else 0)


def run():
    typer.run(main)



import ast

import typer


def echo_with_line_ref(filepath: str, section: ast.AST, message: str):
    filepath = filepath if filepath.startswith("/") else f"./{filepath}"
    typer.echo(f"{filepath}:{section.lineno}:{section.col_offset}: {message}")



from __future__ import annotations

import abc
import ast
import typing as t

from pyaphid.helpers import echo_with_line_ref


class ImportFrom(t.NamedTuple):
    value: ast.ImportFrom
    import_string: str
    alias: str | None


class Import(t.NamedTuple):
    value: ast.Import
    import_string: str
    alias: str | None


class CallMatch(t.NamedTuple):
    call: ast.Call
    match: str


def get_call_signature(call: ast.Call) -> tuple[str, str]:
    if "value" in call.func._fields:
        path = ""
        basename = call.func.attr  # type: ignore
        call_func = call.func.value  # type: ignore
        while "value" in call_func._fields:
            path = f".{call_func.attr}{path}"
            call_func = call_func.value
        if "id" in call_func._fields:
            path = call_func.id + path
        else:
            path = f"{call_func.func.id}(){path}"
    else:
        basename = call.func.id  # type: ignore
        path = ""
    return (path, basename)


def replace_alias(import_: ImportFrom | Import, path: str):
    if import_.alias and (path == import_.alias or path.startswith(import_.alias)):
        return path.replace(import_.alias, import_.import_string, 1)
    return path


def expand_call(call: ast.Call, imports: list[Import], import_froms: list[ImportFrom]):
    """
    Expand call with the matching import string.
    If the call does not match an import and is not a built-in function, None is returned
    """
    (path, basename) = get_call_signature(call)
    if "(" in path or ")" in path:
        return None
    work_path = path
    if path:
        for import_ in reversed(imports):
            work_path = replace_alias(import_, work_path)
            if work_path == import_.import_string or work_path.startswith(
                f"{import_.import_string}."
            ):
                return f"{work_path}.{basename}"
            work_path = path
    for import_from in reversed(import_froms):
        imported_name: str = import_from.import_string.rsplit(".", 1)[1]
        if path:
            work_path = replace_alias(import_from, work_path)
            if work_path == imported_name or work_path.startswith(f"{imported_name}."):
                return f"{work_path.replace(imported_name, import_from.import_string, 1)}.{basename}"  # noqa: E501
        elif (
            import_from.alias
            and basename == import_from.alias
            or imported_name == basename
        ):
            return import_from.import_string
    return basename if not path and basename in __builtins__ else None  # type: ignore


_T = t.TypeVar("_T", bound="ast.FunctionDef | ast.AsyncFunctionDef | ast.ClassDef")


class ImportsTracker(metaclass=abc.ABCMeta):
    def __init__(self) -> None:
        self.import_froms: list[ImportFrom] = []
        self.imports: list[Import] = []

    @abc.abstractmethod
    def generic_visit(self, node: ast.AST) -> t.Any:
        pass

    def new_import_context(self, node: _T) -> _T:
        old_imports = self.imports.copy()
        old_import_froms = self.import_froms.copy()
        self.generic_visit(node)
        self.imports = old_imports
        self.import_froms = old_import_froms
        return node

    def visit_FunctionDef(self, node: ast.FunctionDef) -> ast.FunctionDef:
        return self.new_import_context(node)

    def visit_AsyncFunctionDef(
        self, node: ast.AsyncFunctionDef
    ) -> ast.AsyncFunctionDef:
        return self.new_import_context(node)

    def visit_ClassDef(self, node: ast.ClassDef) -> ast.ClassDef:
        return self.new_import_context(node)

    def visit_Import(self, node: ast.Import) -> ast.Import:
        for name in node.names:
            self.imports.append(Import(node, name.name, name.asname))
        return node

    def visit_ImportFrom(self, node: ast.ImportFrom) -> ast.ImportFrom:
        for name in node.names:
            self.import_froms.append(
                ImportFrom(node, f"{node.module}.{name.name}", name.asname)
            )
        return node


class ExpandedCallCollector(ast.NodeVisitor, ImportsTracker):
    def __init__(self, *args, **kw) -> None:
        self.calls: list[CallMatch] = []
        return super().__init__()

    def visit_Call(self, node: ast.Call) -> None:
        expanded_call_signature = expand_call(node, self.imports, self.import_froms)
        if expanded_call_signature and expanded_call_signature not in map(
            lambda call: call.match, self.calls
        ):
            self.calls.append(CallMatch(node, expanded_call_signature))


class VisitorMixIn(ImportsTracker):
    def __init__(self, filepath: str, forbidden: list[str]) -> None:
        self.filepath = filepath
        self.forbidden = forbidden
        self.matches: list[CallMatch] = []
        return super().__init__()

    def is_forbidden(self, signature: str):
        for pattern in self.forbidden:
            if pattern.endswith(".*") and signature.startswith(
                pattern.rsplit(".", 1)[0]
            ):
                return True
            elif signature == pattern:
                return True
        return False

    def new_import_context(self, node: _T) -> _T:
        if node.name in self.forbidden:
            self.forbidden.remove(node.name)
            echo_with_line_ref(
                self.filepath,
                node,
                f"Local definition  of {node.name} collides with forbidden built-in. {node.name} calls will be ignored for the rest of the file",  # noqa: E501
            )
        return super().new_import_context(node)

    def visit_Call(self, node: ast.Call) -> ast.Call | None:
        expanded_call_signature = expand_call(node, self.imports, self.import_froms)
        if expanded_call_signature and self.is_forbidden(expanded_call_signature):
            self.matches.append(CallMatch(node, expanded_call_signature))
            return None
        return node


class Visitor(ast.NodeVisitor, VisitorMixIn):
    pass


class Transformer(ast.NodeTransformer, VisitorMixIn):
    pass



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pyaphid']

package_data = \
{'': ['*']}

install_requires = \
['tomli>=2.0.1,<3.0.0', 'typer[all]>=0.6.1,<0.7.0']

entry_points = \
{'console_scripts': ['pyaphid = pyaphid.cli:run']}

setup_kwargs = {
    'name': 'pyaphid',
    'version': '0.1.1',
    'description': 'Identify debugging function calls',
    'long_description': '# Pyaphid (Experimental)\n\nPyaphid is a tool for detecting unwanted function calls in Python code.\n\n## Installation and usage\n\nInstallation: `pip install pyaphid`\n\nUsage: `python -m pyaphid <files and/or directories to analyze>` or `pyaphid <files and/or directories to analyze>`\n\n### Configuration\n\nForbidden function calls can be configured via the `pyproject.toml`:\n\n```toml\n[tool.pyaphid]\nforbidden = [\n    "print",\n    "pdb.run",\n    "werkzeug.debug.*"\n]\n```\n\n### CLI Options\n\n- -n / --names: `Look-up all func calls and print their identifier`\n\n## Limitations\n\n```python\n# Pyaphid cannot work with star imports\nfrom os.path import *\ndirname(".") # undetected\n\n# Pyaphid doesn\'t track assignments\nmy_print = print\nmy_print("Hello world") # undetected\n```\n',
    'author': 'Jan Vollmer',
    'author_email': 'jan@vllmr.dev',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/jvllmr/pyaphid',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
